/*
								+---------------------------------+
								|                                 |
								|  ***   Tiny jpeg decoder   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|             Jpeg.h              |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 08-30-2006
 *
 *	write code for __JPEG_DEC_ALLOW_SUBSAMPLED_IMAGES code path (not really necessary)
 *	write code to support 12-bit extended DCT images
 *	refine code for lossless compression to eliminate rounding error when decompressing > 8bit
 *	lossless images (alloc just two int16 scanlines to remember predictor values)
 *
 */

#ifndef __BMP_INCLUDED
#define __BMP_INCLUDED

struct TBmp {
	char n_former_bpp; // former bpp, before conversion to RGBA8
	bool b_grayscale;
	bool b_alpha;
	int n_width, n_height;
	unsigned __int32 *p_buffer;
};

#endif //__BMP_INCLUDED

#ifndef __TINY_JPEG_INCLUDED
#define __TINY_JPEG_INCLUDED

/*
 *		=== config section ===
 */

#define __JPEG_DEC_MAX_COMPONENT_NUM 4
// maximal component number. decoding routines are written to support arbitrary number of
// components (colorspace conversion routines support 1 - 4 components), almost all
// implementatinos use 4, maximum possible component count in jpeg is 255

//#define __JPEG_DEC_VERBOSE
// if defined, prints tables, header info, ... (to stdout)

//#define __JPEG_DEC_READ_VERBOSE
// if defined, prints some debug info about file input (to stdout)

//#define __JPEG_DEC_DECODE_VERBOSE
// if defined, prints decoding progress (i.e. huffman codes, DCT coeffs; to stdout)

//#define __JPEG_DEC_ALLOW_SUBSAMPLED_GRAYSCALE
// jpeg specs allow grayscale images with sampling 1x1 only. if defined, decoder is able to
// render grayscale images with other sampling factors as well (do not affect rendering of
// color images)

//#define __JPEG_DEC_SUPPORT_LOSSLESS
// if defined, decoder can handle lossless jpeg images as well

#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
#define __JPEG_DEC_USE_LOSSLESS_NONSTD_RLE
#endif
// if defined, jpeg lossless algorithm is further enhanced by RLE compression this is non-standard
// extension which requires all images to be written by encoder, supporting this extension
// never use in programs where your jpeg files are available in standalone-file form so they could
// be potentialy copied away! do not breed nor allow anyone to breed non-standard jpegs away from
// your app (note there is no record in image file header wheter this extension was used or not)

#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
//#define __JPEG_DEC_ALLOW_SUBSAMPLED_IMAGES
// do not use! there's no code to support it yet (todo - next version)
#endif
// if defined, decoder can decode lossless images with subsampling other than 1x1
// (it's a bit useless thing to bother with lossless compression and then downsample the
// image; moreover it's not very logical to do colorspace transformation in lossless mode
// as that would introduce more error into image so unless you store your images in YCbCr(K)
// format, subsampling is useless as well ... but despite it's foolish, it's possible)

//#define __JPEG_DEC_SUPPORT_RG
// enable branch that can unpack RG images (i use them for uv-maps like water distortion, etc)
// another practice on how to interpret two-channel jpeg images is 16-bit RGB (acd-see)

//#define __JPEG_DEC_SUPPORT_YCCK
// enable branch that can unpack YCCK images

#ifndef __JPEG_DEC_SUPPORT_YCCK
//#define __JPEG_DEC_SUPPORT_CMYK
#endif
// enable branch that can unpack CMYK images (jpeg does contain component number only,
// not component meaning so it's possible to handle 4-component jpegs either as YCCK (common)
// or as CMYK (possible, but not widely used))

//#define __JPEG_DEC_INTEGER_IDCT
// if defined, use faster integer IDCT. otherwise use precise, slower float IDCT
// (in msvc, it's faster only in optimize speed mode. in optimize size, it's a little bit slower)

#ifdef __JPEG_DEC_INTEGER_IDCT
#define __JPEG_DEC_USE_MMX
#endif
// if defined, faster MMX integer IDCT path is used, note code is a bit larger (~500 bytes)

#define __JPEG_DEC_STRIP_ERROR_CHECKS
// if defined, all error checks aparts from alloc success tests are removed - saves a few code bytes
// (arround 1 kB; use with precaution - errors during decoding process may make your program crash)

#define __JPEG_DEC_BGR
// if defined, swap R and B during colorspace transformation

/*
 *		=== ~config section ===
 */

class CJpegDecoder {
protected:
	static char p_zig_indices[/*64*/63]; // don't need the first one; actualy inverse zig sequence

	class CFileBuffer { // class for fast buffered file input (also assures byte-order reversion)
	protected:
		unsigned char m_p_buffer[16384];
		static const int m_n_buffer_size;
		int m_n_buffer_used;
		unsigned char *m_p_buffer_ptr;
		FILE *m_p_fr;

	public:
		CFileBuffer(FILE *p_fr);
		bool __fastcall Read_Byte(unsigned char &r_n_byte);
		bool __fastcall Read_Short(short &r_n_short);
		bool inline Read_Short(unsigned short &r_n_short) { return Read_Short((short&)r_n_short); }
		bool __fastcall Read_Marker(unsigned short &r_n_marker_code);
		bool __fastcall Skip_Bytes(int n_byte_num);
	};

	class CBitReader { // class for reading bit-by-bit, used in huffman decoder
	protected:
		unsigned char m_n_cur_byte; // current byte, it's shifted as we eat bits
		bool m_b_0xff_byte; // set to true if last byte read was 0xff
		unsigned char m_n_bit_num; // number of valid bits in byte
		bool m_b_error_flag;

	public:
		CBitReader();
		bool inline b_Error() { return m_b_error_flag; }
		unsigned char __fastcall n_GetBit(CFileBuffer *p_file_buffer);
		short __fastcall n_GetNumber(int n_bit_num, CFileBuffer *p_file_buffer);
		void Restart();
	};

	struct TBlockAction { // table for block actions
		unsigned short n_id;
		bool (*p_function)(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	};
	static const TBlockAction m_p_action_table[];

	struct TQuantTable { // quantization table
		bool b_16_bit; // otherwise 8-bit
#ifdef __JPEG_DEC_INTEGER_IDCT
		int p_value[64]; // 16:16 fixed-point (short would be enough for 8-bit quant tables, but not for 16 bit)
#else
		float p_value[64]; // hack! - should be unsigned short, but i pre-multiplied it and can use faster IDCT function
#endif
	} m_p_quant_table[4];
	// quantization tables (can be separate one for each component)

	struct THuffmanTable {
		bool b_valid; // did we read it from the file? in case we didn't, it could make our program crash, (unlikely if quantization tables are missing, image just looks scrambled but it can never make program crash)
		unsigned char p_code_num[16]; // number of codes for all possible 16 lengths
		unsigned char *p_code[16]; // 16 pointers into p_code_table where the actual data are
		unsigned short p_min_code_value[16]; // min and max code value; n+1 bit code must be >= p_min_code_value[n] && < p_max_code_value[n] (thus in case p_code_num[n] == 0 then also p_min_code_value[n] == p_max_code_value[n])
		//unsigned short p_max_code_value[16]; // don't store this, it's equal to p_min_code_value[n] + p_code_num[n]
		unsigned char p_code_table[256]; // table for all huffman codes
	} m_p_huffman_table[4][2]; // 0 = AC, 1 = DC
	// huffman tables (can be separate one for each component)

	struct TComponentInfo {
		unsigned char n_component_id; // 1 = Y, 2 = Cb, 3 = Cr (JFIF standard)
		unsigned char n_sampling_horiz;
		unsigned char n_sampling_vert; // component sampling (1, 2, or 4 usually)
		unsigned char n_scale_horiz;
		unsigned char n_scale_vert;
		unsigned char n_quant_table_index; // used quantization table
	};
	// outside TImageHeader because of scope issues (needed visibility for TScanHeader::TComponentInfo)

	struct TImageHeader {
		unsigned short n_frame_start_marker; // marker_SOF0_HuffBaseline_DCT supported for now only
		unsigned char n_sample_precision; // 8 (8 or 12 for extended baseline or progressive)
		short n_height;
		short n_width;
		int n_macro_block_num, n_macro_block_x_num; // number of macro-blocks
		unsigned char n_component_num; // should be 1-4 (1 = gray, 3 = YCbCr, 4 = CMYK)
		TComponentInfo p_component_info[__JPEG_DEC_MAX_COMPONENT_NUM];
	} m_t_image_header;
	// image header

	struct TRestartInterval {
		bool b_enabled;
		unsigned short n_interval;
	} m_t_restart;
	// restart interval definition

	struct TScanHeader {
		unsigned char n_component_num; // should match TImageHeader::n_component_num

		struct TComponentInfo2 {
			unsigned char n_component_id; // should match TImageHeader::TComponentInfo::n_component_id
			TComponentInfo *p_component; // points to image header component
			unsigned char n_DC_entropy_table;
			unsigned char n_AC_entropy_table; // table indices <0, 3>
			THuffmanTable *p_DC_table, *p_AC_table;
		} p_component_info[__JPEG_DEC_MAX_COMPONENT_NUM];

		unsigned char n_spectrum_first; // index (in zig-zag order) of first coded coefficient
		unsigned char n_spectrum_last; // index (in zig-zag order) of last coded coefficient (should be set to 63 for sequential DCT)
		unsigned char n_sucessive_approximation; // two nibbles, point transform, should be zero (not used in marker_SOF0_HuffBaseline_DCT)
	};

	struct TImageData {
		bool b_read_image_header;
#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
		bool b_lossless; // set to true in case marker_SOF3_HuffLossless was read; valid only if b_read_image_header == true
#endif
		unsigned char *p_framebuffer[__JPEG_DEC_MAX_COMPONENT_NUM]; // frame-buffers, as big as image size / component sampling
		unsigned short p_framebuffer_width[__JPEG_DEC_MAX_COMPONENT_NUM];
		unsigned short p_framebuffer_height[__JPEG_DEC_MAX_COMPONENT_NUM];
		short p_dc_value[__JPEG_DEC_MAX_COMPONENT_NUM]; // dc accumulators
	} m_t_image_data; // cleared always when starting decoding jpeg

	template <const int n_component_num, class TColorConversion>
	class CImageRenderer;
	// template body in Jpeg.cpp (don't want to put it here as it's pretty useless)

public:
	/*
	 *	TBmp *p_Decode_JPEG(FILE *p_fr)
	 *		- decode jpeg, convert it to RGBA, 8 bits per channel bitmap
	 *		- alpha is always 0xff since jpeg files don't store alpha channel
	 *		- return 0 on failure (not enough memory / unsupported format / damaged file)
	 */
	TBmp *p_Decode_JPEG(FILE *p_fr);

	/*
	 *	void Supply_QuantTable(int n_index, const void *p_table, bool b_16_bit)
	 *		- it's possible to supply quantization tables before decoding image
	 *		- (image is decoded anyway, even if it doesn't contain quantization
	 *		  tables and you can save some space if all your images use the same
	 *		  quantization tables)
	 *		- n_index has to be in range 0-3 (there are four tables in jpeg)
	 *		- p_table points to table data (if b_16_bit == false, then it's array
	 *		  of 64 unsigned chars, if b_16_bit == true then it's array of 64
	 *		  unsigned shorts)
	 */
	bool Supply_QuantTable(int n_index, const void *p_table, bool b_16_bit);

protected:
	static bool Read_SkipMarker(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_Unsupported(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_Skip(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_JFIF_Header(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_RestartInterval(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_QuantTables(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_HuffmanTables(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_FrameHeader(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	static bool Read_StartScan(unsigned short n_block_id, CJpegDecoder *p_this, CFileBuffer *p_file_buffer);
	bool Decode_Baseline_MacroBlock(int n_macro_block_index, const TScanHeader *p_scan_head,
		CBitReader &r_bit_reader, CFileBuffer *p_file_buffer);
#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
	bool Decode_Lossless_Image(const TScanHeader *p_scan_head,
		CBitReader &r_bit_reader, CFileBuffer *p_file_buffer);
#endif
};

#endif //__TINY_JPEG_INCLUDED
